﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter {
  using System;
  using HeuristicLab.Core;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Configuration;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Data.Pool;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;
  using HeuristicLab.Random;

  public class PushInterpreterPool {
    private readonly ObjectPool<PooledPushInterpreter> pool;

    public readonly ManagedPoolProvider<PushProgram> PushProgramPoolProvider;
    public readonly ManagedPoolProvider<LoopState> LoopStatePoolProvider;
    public readonly ManagedPoolProvider<PooledList<Expression>> ExpressionListPoolProvider;


    public PushInterpreterPool(IReadOnlyPushConfiguration config = null)
      : this(Environment.ProcessorCount * 2, 1024, null, config) {
    }

    public PushInterpreterPool(int size, int pushProgramPoolPartitionSize, int? maxPartitionCount = null, IReadOnlyPushConfiguration config = null) {
      PushGpConfiguration = config ?? new PushConfiguration();

      PushProgramPoolProvider = new ManagedPoolProvider<PushProgram>(pushProgramPoolPartitionSize, () => new PushProgram(), maxPartitionCount);
      LoopStatePoolProvider = new ManagedPoolProvider<LoopState>(pushProgramPoolPartitionSize, () => new LoopState(), maxPartitionCount);
      ExpressionListPoolProvider = new ManagedPoolProvider<PooledList<Expression>>(pushProgramPoolPartitionSize * 2, () => new PooledList<Expression>(), maxPartitionCount);

      pool = new ObjectPool<PooledPushInterpreter>(() => {
        var poolContainer = new InterpreterPoolContainer(PushProgramPoolProvider, LoopStatePoolProvider, ExpressionListPoolProvider);
        return new PooledPushInterpreter(this, PushGpConfiguration, poolContainer);
      }, size);
    }

    public IReadOnlyPushConfiguration PushGpConfiguration { get; private set; }

    public PooledPushInterpreter Create(IRandom random = null) {
      var interpreter = this.pool.Allocate();
      interpreter.Random = random ?? new FastRandom();

      return interpreter;
    }

    public void Free(PooledPushInterpreter interpreter) {
      this.pool.Free(interpreter);
    }
  }
}
